package Sub::Scheduler;

use version; $VERSION = qv('0.0.4'); # update POD and Changes!

use warnings;
use strict;
use Carp;

use Perl6::Export::Attrs;
use Time::HiRes     qw( time );
use List::MoreUtils qw( first_index );

my $next_taskid = 0;
my @tasks;

# OPTIMIZATION
#    @tasks     start_sec .
# +  run_tasks()      N  @tasks.
# -  add_task()        @tasks, 
#     -   start_sec.
# -  run_tasks()         @tasks
#       ţ  repeat_sec.
# *  del_task()    .

sub _add_task {
    my ($task) = @_;
    my $idx = first_index { $_->{start_sec} >= $task->{start_sec} } @tasks;
    $idx = @tasks if $idx == -1;
    splice @tasks, $idx, 0, $task;
}

sub add_task :Export(:MANDATORY) {
    my ($sub_ref, $start_sec, $repeat_sec) = @_;
    croak 'usage: add_task(\&sub, $start_sec[, $repeat_sec])'
        if ref $sub_ref ne 'CODE' 
        || !defined $start_sec 
        || $start_sec <= 0
        || (defined $repeat_sec && $repeat_sec <= 0);
    #   .
    $next_taskid++;
    my $new_task = {
        sub_ref     => $sub_ref,
        start_sec   => $start_sec,
        repeat_sec  => $repeat_sec,
        taskid      => $next_taskid,
    };
    #     .
    _add_task($new_task);
    #    .
    return $next_taskid;
}

sub del_task :Export(:MANDATORY) {
    my ($taskid) = @_;
    my $idx = first_index { $_->{taskid} == $taskid } @tasks;
    croak 'del_task: no such task' if $idx == -1;
    splice @tasks, $idx, 1;
}

sub run_tasks :Export(:MANDATORY) {
    my $time = shift || time();
    return unless @tasks;   # OPTIMIZATION
    my $exception;
    #  ,    (     ):
    my $late_idx = first_index { $_->{start_sec} > $time } @tasks;
    $late_idx = @tasks if $late_idx == -1;
    TASK:
    for my $task (splice @tasks, 0, $late_idx) {
        #       , 
        #        run_tasks.
        if ($exception) {
            _add_task($task);
            next TASK;
        }
        #  .
        eval { $task->{sub_ref}->(); };
        $exception=$@ if $@;
        #    ,   ţ start_sec 
        #   .
        next TASK if !defined $task->{repeat_sec};
        $task->{start_sec} = $time + $task->{repeat_sec};
        _add_task($task);
    }
    #         - rethrow !
    die $exception if $exception;
}

sub get_next_run_time :Export(:MANDATORY) {
    return undef unless @tasks;
    return $tasks[0]->{start_sec};
}

1; # Magic true value required at end of module
__END__

=head1 NAME

Sub::Scheduler - Schedule sub execution at given time.


=head1 VERSION

This document describes Sub::Scheduler version 0.0.4


=head1 SYNOPSIS

    use Sub::Scheduler;

    add_task(\&sub, $start_sec);
    $taskid = add_task(\&sub, $start_sec, $repeat_sec);
    $sec = get_next_run_time();
    sleep $sec-time();
    run_tasks();
    del_task($taskid);

=head1 DESCRIPTION

Postpone sub execution until given time (with possible repeated executions
every X seconds). To execute these subs user should manually execute
run_tasks() from time to time (i.e. within "main loop" of user's code).

All time should be given in seconds, possibly with fractional part
(i.e. same as Time::HiRes format).

=head1 INTERFACE 

=over

=item add_task( \&sub, $start_sec )

=item add_task( \&sub, $start_sec, $repeat_sec )

Schedule \&sub to be executed at $start_sec. If $repeat_sec given then
\&sub will be executed again after $repeat_sec after previous execution
until user stop it with del_task().

Return task identifier (scalar) which can be used to stop this task with
del_task().

=item del_task( $taskid )

Stop task $taskid ($taskid is value returned by add_task()).

If no such task exists (because this task was already deleted by del_task()
OR this task has no $repeat_sec and was already executed and deleted by
run_tasks() OR $taskid is just wrong) - 
throw exception 'del_task: no such task'.

=item run_tasks( )

=item run_tasks( $time )

Run all scheduled tasks which should be started before now (or $time, if given).

If task doesn't have $repeat_sec, then delete this task after execution.
If task have $repeat_sec, then reschedule this task to now (or $time) plus
$repeat_sec.

If task throw exception then no other tasks planned for current run_tasks()
execution will be runned and that exception will be rethrown by run_tasks().

=item get_next_run_time( )

Return next time when some task should be executed. Useful to sleep() or
do some operation with that timeout before execute run_tasks().

=back

=head1 DIAGNOSTICS

=over

=item C<< usage: add_task(\&sub, $start_sec[, $repeat_sec]) >>

You've executed add_task() with wrong params.

=item C<< del_task: no such task >>

You try to delete task using wrong $taskid. See del_task() description above
for complete list of cases when this may happens.

=back


=head1 CONFIGURATION AND ENVIRONMENT

Sub::Scheduler requires no configuration files or environment variables.


=head1 DEPENDENCIES

None.


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to
C<bug-sub-scheduler@rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org>.

=head1 AUTHOR

Alex Efros  C<< <powerman@asdfgroup.com> >>

Ilya Persky  C<< <ilya_persky@yahoo.com> >>

=head1 LICENCE AND COPYRIGHT

Copyright (c) 2005, Alex Efros C<< <powerman@asdfgroup.com> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
